#!/usr/bin/python3
import os
import time
import requests
import cv2
import magic
import numpy as np

#Go here for api key:
#https://azure.microsoft.com/en-us/try/cognitive-services/
#Search apis -> Bing Search APIs v7 -> API key should be at the bottom, looks like '0A1B2C3D4E...'
apiKey = ''
url = 'https://api.cognitive.microsoft.com/bing/v7.0/images/search'

#search request variables
pageSize = 20
nPages = 10

downloadPath = './images'

#one in every x images is segregated for evaluation purposes
testRatio = 10

#desired image dimensions, to match model input
imageDimensions = (224, 224)

#'magic' library is used to determine file mime types
m = magic.Magic(mime=True)

#python2 compatibility
try:
	input = raw_input
except NameError:
	pass

def getSearchResults(searchTerm, pageSize, pages):

	json = []

	for i in range(0, pages):

		#create request data for microsofts api
		headers = {"Ocp-Apim-Subscription-Key" : apiKey}
		params = {"q": searchTerm, "offset": i * pageSize, "count": pageSize}

		#perform search, rate limit to 3 requests per second for microsofts anti-spam
		search = requests.get(url, headers=headers, params=params)
		search.raise_for_status()
		json.append(search.json())
		time.sleep(0.34)

	return json

def main():

	#prompt the user for search term
	nResults = str(pageSize * nPages)
	searchTerm = input('Choose a search term to download: ')
	print("Attempting to get search results for " + nResults + " pictures of " + searchTerm)

	#get search results
	try:
		searchData = getSearchResults(searchTerm, pageSize, nPages)
	except:
		print("Error connecting to bing, check internet connection, api key, subscription etc...")
		return

	print("Creating directory for results")
	trainingPath = os.path.join(downloadPath, "training", searchTerm)
	testingPath = os.path.join(downloadPath, "testing", searchTerm)

	try:
		os.makedirs(trainingPath)
		os.makedirs(testingPath)
	except OSError:
		print("Unable to create directories - possibly already exist")

	print("Grabbing full res images from search results")

	nextTrainingFilename = 1
	nextTestingFilename = 1

	for page in searchData:
		for image in page["value"]:
			try:
				#attempt to download image. May fail for permissions etc, 
				#If so we dont care and can just move on to the next

				req = requests.get(image["contentUrl"], timeout=4)

				#confirm mime manually to verify integrity, and confirm common format jpeg, png
				mime = m.from_buffer(req.content)
				if mime != "image/jpeg" and mime != "image/png":
					continue

				#decode image from buffer into opencv, force 3 channels (RGB)
				imgArray = np.frombuffer(req.content, dtype=np.uint8)
				img = cv2.imdecode(imgArray, 1)

				#attempt to resize using opencv
				#remembering opencv flips x,y, amongst other things (bgr..?)
				img = cv2.resize(img, (imageDimensions[1], imageDimensions[0]))

				#write resized image to either test or training folder
				fileExtension = "." + mime.split("/")[1]

				if (nextTrainingFilename + nextTestingFilename) % testRatio == 0:
					cv2.imwrite(os.path.join(testingPath, str(nextTestingFilename) + fileExtension), img)
					nextTestingFilename += 1
				else:
					cv2.imwrite(os.path.join(trainingPath, str(nextTrainingFilename) + fileExtension), img)
					nextTrainingFilename += 1

				if (nextTestingFilename + nextTrainingFilename) % 100 == 0:
					print("Attempted " + str(nextTestingFilename + nextTrainingFilename - 2) + " images")
			except:
				pass

	print("Finished downloading images of " + searchTerm)
	print("Managed to download " + str(nextTestingFilename + nextTrainingFilename - 2) + " images out of " + nResults + " attempted")

if __name__ == '__main__':
	main()
