from keras.preprocessing.image import ImageDataGenerator
from keras.optimizers import Adam
from keras.preprocessing.image import img_to_array
from keras.models import load_model
from keras import backend as K

from sklearn.preprocessing import LabelBinarizer
from sklearn.model_selection import train_test_split

import matplotlib.pyplot as plt
import numpy as np

import vgg
import pickle
import cv2
import os
import math

try:
	import PIL
except:
	sys.path.append(".")

datasetPath = './images'
modelPath = "./models"

modelSize = (224, 224, 3)

#training parameters
epochs = 15
initialLearningRate = 0.001
batchSize = 8

#python2 compatibility
try:
	input = raw_input
except NameError:
	pass

def getLabels():

	labels = []

	for i in os.walk(datasetPath):
		if i[0] == (os.path.join(datasetPath, "training")):
			labels = i[1]

	labels = np.array(labels[::-1])
	lb = LabelBinarizer()
	labels = lb.fit_transform(labels)

	return lb

def getModel(labels):

	global modelSize

	isNewModel = input("Create a new model or retrain an old one? (N/O) ").lower()

	if isNewModel == "o":

		existingModels = []

		for i in os.walk(modelPath):
			if i[0] == modelPath:
				existingModels = i[1]

		print("Existing Models: ")
		print(existingModels)

		modelName = input("Please enter the model name: ")

		if modelName in existingModels:

			model = load_model(os.path.join("models", modelName, "model"))
			return model, modelName
		else:
			print("Unable to find model: " + modelName)
			return None, None

	elif isNewModel == "n":

		modelName = input("Enter name for new model: ")
		model = vgg.buildVGG(modelSize[0], modelSize[1], modelSize[2], len(labels.classes_))

		return model, modelName

	else:

		print("Unrecognised command")
		return None, None

def getDatasetStats(datasetPath, batchSize):

	nTrainSamples = 0
	for i in os.walk(os.path.join(datasetPath, "training")):
		nTrainSamples += len(i[2])

	nTestSamples = 0
	for i in os.walk(os.path.join(datasetPath, "testing")):
		nTestSamples += len(i[2])

	stepsPerEpoch = nTrainSamples / float(batchSize)
	validationSteps = nTestSamples / float(batchSize)

	if nTrainSamples == 0 or nTestSamples == 0:
		print("Error finding training images. Looking in " + datasetPath + "/training or /testing")

	else:
		print("Steps per epoch: " + str(math.ceil(stepsPerEpoch)))
		print("Validation steps: " + str(math.ceil(validationSteps)))

	return stepsPerEpoch, validationSteps

def train(modelSize, batchSize, model, initialLearningRate, epochs):

	stepsPerEpoch, validationSteps = getDatasetStats(datasetPath, batchSize)

	trainingGenerator = ImageDataGenerator(
		rescale = 1./255,
		rotation_range = 25,
		width_shift_range = 0.1,
		height_shift_range = 0.1, 
		shear_range = 0.2,
		zoom_range = 0.2,
		horizontal_flip = True,
		fill_mode = "nearest")
	testingGenerator = ImageDataGenerator(
		rescale = 1,
		rotation_range = 2,
		width_shift_range = 0.1,
		height_shift_range = 0.1, 
		shear_range = 0.1,
		zoom_range = 0.2,
		horizontal_flip = True,
		fill_mode = "nearest")

	trainingImageGenerator = trainingGenerator.flow_from_directory(
		os.path.join(datasetPath, "training"),
		target_size = (modelSize[0], modelSize[1]),
		batch_size = batchSize,
		color_mode = 'rgb',
		class_mode = 'categorical')

	testingImageGenerator = testingGenerator.flow_from_directory(
		os.path.join(datasetPath, "testing"),
		target_size = (modelSize[0], modelSize[1]),
		batch_size = batchSize,
		color_mode = 'rgb',
		class_mode = 'categorical')

	print("Compiling neural network")
	optimser = Adam(lr=initialLearningRate, decay=initialLearningRate/epochs)
	model.compile(loss="categorical_crossentropy", optimizer=optimser, metrics=["accuracy"])

	print("Training network")
	history = model.fit_generator(
		trainingImageGenerator,
		steps_per_epoch = stepsPerEpoch,
		epochs = epochs,
		validation_data = testingImageGenerator,
		validation_steps = validationSteps,
		verbose = 1)
	print("Finished training")

	return history

def saveModel(model, labelEncoder, modelName, history):

	print("Saving model")
	filename = os.path.join(modelPath, modelName, "model")
	model.save(filename)

	print("Saving Labels")
	filename = os.path.join(modelPath, modelName, "labels")
	with open(filename, "wb") as f:
		pickle.dump(labelEncoder, f)

	print("Saving/appending training history")
	filename = os.path.join(modelPath, modelName, "history")

	#if the history file already exists, merge together the old and new, else just write to file
	if os.path.isfile(filename):

		prevHistory = None
		with open(filename, "rb") as f:
			prevHistory = pickle.load(f)

		newHistory = {}
		for key in prevHistory:
			newKey = prevHistory[key] + history.history[key]
			newHistory[key] = newKey

		with open(filename, "wb") as f:
			pickle.dump(newHistory, f)

	else:
		with open(filename, "wb") as f:
			pickle.dump(history.history, f)

	print("Finished saving experiment")

def createPlot(epochs, history, modelName):

	print("Saving plot of training")

	plt.style.use("seaborn-talk")
	plt.figure()

	plt.plot(range(0, epochs), history.history["loss"], label="Training Loss")
	plt.plot(range(0, epochs), history.history["acc"], label="Training Accuracy")
	plt.plot(range(0, epochs), history.history["val_loss"], label="Validation Loss")
	plt.plot(range(0, epochs), history.history["val_acc"], label="Validation Accuracy")

	plt.title("Training History")
	plt.xlabel("Epoch")
	plt.ylabel("Value")
	plt.legend(loc="upper right")

	plt.savefig(os.path.join(modelPath, modelName, "history.png"))

	print("Plot saved")

def main():

	global modelSize, epochs, initialLearningRate, batchSize

	#get image labels and model
	lb = getLabels()
	model, modelName = getModel(lb)

	if model == None:
		return

	try:
		os.makedirs(os.path.join("models", modelName))
	except OSError:
		print("Unable to create directories - possibly already exist")

	#train model
	history = train(modelSize, batchSize, model, initialLearningRate, epochs)

	#save results
	saveModel(model, lb, modelName, history)
	createPlot(epochs, history, modelName)

if __name__ == '__main__':
	main()
